package generators;

import terms.*;
import parsers.*;
import util.*;

/** A very simple implementation of deterministic bottom-up tree transducers.
  * The basic behaviour of an object of this class can be controlled by the commands
  * <code>single steps</code> and <code>complete runs</code>. Normally, if an input
  * term is given to the transducer, rules will be applied as long as possible and, upon
  * successful termination, the resulting term is returned (otherwise, the result is
  * <code>null</code>). However, using <code>single steps</code> one can switch to
  * a mode where single derivation steps are applied in a stepwise manner. In this
  * case, there is an additional command <code>next step</code> which acts in the
  * obvious way. Using the command <code>complete runs</code> one can switch back
  * to the default behaviour.
  */
public class dbuTransducer extends treeTransducer {

//====================================================================
// First the 'interactive' part that deals with commands ...
//====================================================================
  private static String nextStep = "single step";
  private static String back = "back";
  private static String singleSteps = "derive stepwise";
  private static String completeRuns = "results only";
  private boolean stepwise = false;
  
/** Returns the (names of the) commands described above. */
  public list commands() {
    list result = new list();
    String[] com;
    if (stepwise) {
      com = new String[2];
      com[0] = nextStep;
      com[1] = back;
      result.append(com);
    }
    com = new String[1];
    com[0] = stepwise ? completeRuns : singleSteps;
    result.append(com);
    return result;
  }
  
  public void execute(String command) {
    if (nextStep.equals(command)) singleStep();
    else if (back.equals(command)) back();
    else if (singleSteps.equals(command) || completeRuns.equals(command)) {
	    stepwise = !stepwise;
	    stepCount = 0;
    }
  }

//====================================================================
// Now the actual implementation of dbu tree transducers ...
//====================================================================

  private static term dummyTerm = new term(new symbol("dummy",0));
  private term[] lhs;
  private term[] rhs;
  private fixedRankSignature states;
  private finiteSignature finalStates;
  private term lastArgument = null;
  private term ct;
  private symbol antiState = new symbol("AS",1);
  private list redices = new list();
  private int stepCount = 0;
  
/** Compute the output for a given input term. */
  public term apply(term t) {
    lastArgument = t;
    stepCount = 0;
    run();
    return currentTerm();
  }
  
  private void run() {
    extendedState.reset();
    if (lastArgument != null) ct = (term)lastArgument.clone();
    else {
      ct = null;
      return;
    }
    findRedices();
    while (!redices.isEmpty()) {
      term red = (term)redices.head();
      int i = match(red);
      if (i >= 0) replace(red, (term)rhs[i]);
      else insertAntiState(red);
      redices = redices.tail();
      if (redices.isEmpty()) findRedices();
    }
  }

  private void singleStep() {
    if (stepCount <= extendedState.maxCount) stepCount++;
  }
  
  private void back() {
    if (stepCount > 0) stepCount--;
  }
  
  public term currentTerm() {
    if (ct == null) return null;
    if (stepwise && (!antiState.equals(ct.topSymbol()) || stepCount <= extendedState.maxCount)) return mkCurrent(ct);
    else if (finalStates.contains(ct.topSymbol())) return ct.subterm(0);
    else return null;
  }
  
  private term mkCurrent(term ct) {
    symbol top = ct.topSymbol();
    if (top instanceof extendedState) {
      extendedState ext = (extendedState)top;
      if (ext.count < stepCount) return ct;
      else return mkCurrent(ext.savedComputation);
    }
    else {
      term result = new term(top);
      for (int i = 0; i < top.rank(); i++) result.defineSubterm(i, mkCurrent(ct.subterm(i)));
      return result;
    }
  }
  
  private list redEnd;
  
  private void findRedices() {
    redices = new list();
    if (!states.contains(ct.topSymbol())) {
      redEnd = redices;
      findRedices(ct);
    }
  }
  
  private void findRedices(term sct) {
    int r = sct.topSymbol().rank();
    boolean isRedex = true;
    for (int i = 0; i < r; i++) {
      term ssct = sct.subterm(i);
      if (!states.contains(ssct.topSymbol())) {
        isRedex = false;
        findRedices(ssct);
      }
    }
    if (isRedex) {
      redEnd.append(sct);
      redEnd = redEnd.tail();
    }
  }
  
  private int match(term redex) {
    for (int i = 0; i < lhs.length; i++) {
      if (lhs[i].match(redex) != null) return i;
    }
    return -1;
  }
  
  private void replace(term red, term t) {
    symbol f = red.topSymbol();
    term redCopy = new term(f);
    term[] matching = new term[f.rank()];
    for (int i = 0; i < f.rank(); i++) {
      redCopy.defineSubterm(i,red.subterm(i));
      matching[i] = red.subterm(i).subterm(0);
    }
    extendedState newTop = new extendedState(t.topSymbol(),redCopy);
    red.relabel(newTop);
    red.defineSubterm(0, copyRhs(t.subterm(0), matching));
  }
    
  private term copyRhs(term t, term[] matching) {
    symbol top = t.topSymbol();
    if (top instanceof variable) {
      return matching[((variable)top).index()-1];
    }
    else {
      term result = new term(top);
      for (int i = 0; i < top.rank(); i++) {
        result.defineSubterm(i, copyRhs(t.subterm(i), matching));
      }
      return result;
    }
  }
  
  private void insertAntiState(term red) {
    symbol f = red.topSymbol();
    term redCopy = new term(f);
    for (int i = 0; i < f.rank(); i++) redCopy.defineSubterm(i,red.subterm(i));
    extendedState newTop = new extendedState(antiState,redCopy,Integer.MAX_VALUE);
    red.relabel(newTop);
    red.defineSubterm(0, dummyTerm);
  }
        
/** Initialize this object by parsing the definition of a dbu tree transducer.
  * The syntax is defined by the class <code>dbuTransducerParser</code>.
  * @see dbuTransducerParser
  * @exception ParseException if an error occurs
  */
  public void parse(ASCII_CharStream stream) throws ParseException {
    dbuTransducerParser parser = new dbuTransducerParser(stream);
    parser.dbuTransducer();
    finalStates = parser.finalStates;
    lhs = new term[parser.lhsV.size()];
    parser.lhsV.copyInto(lhs);
    rhs = new term[parser.rhsV.size()];
    parser.rhsV.copyInto(rhs);
    states = parser.states;
    states.addSymbol(antiState);
  }
  
  private static class extendedState extends symbol {
    
    public static int maxCount = 0;
    
    public term savedComputation;
    public int count;
    
    public extendedState(symbol top, term save, int count) {
      super(top.toString(), top.rank());
      if (top.rank() != 1) throw new InternalError("State of rank != 1 in implementation of dbu Transducers encountered.");
      savedComputation = save;
      this.count = count;
    }
    
    public extendedState(symbol top, term save) {
      this(top,save,maxCount++);
    }
    
    public static void reset() {
      maxCount = 0;
    }
  }
    
}
